#pragma once

// #pragma message "Including the src config.h file from the binary dir."

/* #undef MXE */

#define XPERTMASS_INSTALL_PREFIX "/usr/local"

#define XPERTMASS_INSTALL_LIB_DIR "lib"

#define XPERTMASS_INSTALL_CHEMDATA_DIR "share/xpertmass/data"
#define XPERTMASS_INSTALL_DOC_DIR "share/doc/xpertmass"
