/* BEGIN software license
 *
 * MsXpertSuite - mass spectrometry software suite
 * -----------------------------------------------
 * Copyright 2009--2026 by Filippo Rusconi
 *
 * http://www.msxpertsuite.org
 *
 * This file is part of the MsXpertSuite project.
 *
 * The MsXpertSuite project is the successor of the massXpert project. This
 * project now includes various independent modules:
 *
 * - massXpert, model polymer chemistries and simulate mass spectrometric data;
 * - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * END software license
 */


#pragma once


/////////////////////// Qt includes
#include <QString>
#include <QWidget>
#include <QTextEdit>
#include <QPushButton>
#include <QVBoxLayout>
#include <QApplication>
#include <QLabel>
#include <QTimer>

/////////////////////// pappsomspp includes


/////////////////////// Local includes
#include "MsXpS/export-import-config.h"

namespace MsXpS
{
namespace libXpertMassGui
{


class DECLSPEC TextDetailsWidget: public QWidget
{

  Q_OBJECT

  public:
  explicit TextDetailsWidget(const QFont &font, QWidget *parent = nullptr);
  TextDetailsWidget(const QFont &font,
                    const QString &default_status_bar_message = QString(),
                    QWidget *parent                           = nullptr);
  ~TextDetailsWidget();

  void
  showText(const QString &text, const QPoint &globalMousePos, bool is_html);

  void addButton(QPushButton *button_p, bool prepend = false);

  void forcefullyHide();

  void setDefaultStatusBarMessage(const QString &message);
  void showStatusBarMessage(const QString &message, int time_out = 2000);

  void setPinned(bool pinned);

  void enterEvent(QEnterEvent *event) override;
  void leaveEvent(QEvent *event) override;
  bool eventFilter(QObject *obj, QEvent *event) override;

  private slots:

  void copyTextToClipboard();
  void startHideTimer();
  void hideWidget();

  private:
  QVBoxLayout *mp_mainVerticalLayout         = nullptr;
  QHBoxLayout *mp_buttonsRowHorizontalLayout = nullptr;
  QTextEdit *mp_textEdit     = nullptr;
  QPushButton *mp_copyTextButton             = nullptr;
  QLabel *mp_statusBarLabel                  = nullptr;
  QTimer *mp_hideTimer       = nullptr;
  QTimer *mp_statusBarTimer                  = nullptr;
  QString m_defaultStatusBarMessage;
  bool m_isPinned = false;
};


} // namespace libXpertMassGui
} // namespace MsXpS
